import typing

from django.conf import settings


__all__ = ['configurations']


class HealthConfiguration:
    @property
    def enabled(self) -> bool:
        """
        Health enabled flag
        """
        return getattr(settings, 'HEALTH_CHECKERS_ENABLED', True)

    @property
    def checkers(self) -> typing.Set[str]:
        """
        Health checkers (classes)
        """
        default_checkers = {
            'health.checkers.disk.DiskHealthChecker',
            'health.checkers.database.DatabaseHealthChecker',
            'health.checkers.cache.CacheHealthChecker'
        }
        return getattr(settings, 'HEALTH_CHECKERS', default_checkers)

    @property
    def context_path(self) -> str:
        """
        Health checkers context path
        """
        return getattr(settings, 'HEALTH_CHECKERS_CONTEXT_PATH', '/infra')


configurations = HealthConfiguration()



import typing
import pydantic

from health.consts import Status

__all__ = [
    'ComponentHealth',
    'DiskHealth',
    'DatabaseHealth',
    'CacheHealth'
]


class ComponentHealth(pydantic.BaseModel):
    """
    Base info model
    """
    type: str
    status: Status = Status.ALIVE


class DiskHealth(ComponentHealth, pydantic.BaseModel):
    """
    Disk info model
    """
    type = "DISK"
    total: typing.Optional[int]
    free: typing.Optional[int]
    used: typing.Optional[int]


class DatabaseHealth(ComponentHealth, pydantic.BaseModel):
    """
    Database info model
    """
    type = "DATABASE"
    name: typing.Optional[str]
    uri: typing.Optional[str]
    vendor: typing.Optional[str]


class CacheHealth(ComponentHealth, pydantic.BaseModel):
    """
    Cache info model
    """
    type = "CACHE"
    name: typing.Optional[str]
    backend: typing.Optional[str]



from health.apps import HealthConfig
from health.checkers import HealthChecker
from health.checkers import DisabledHealthChecker
from health.checkers import DiskHealthChecker
from health.checkers import DatabaseHealthChecker
from health.checkers import CacheHealthChecker

__all__ = [
    'HealthConfig',
    'HealthChecker',
    'DisabledHealthChecker',
    'DiskHealthChecker',
    'DatabaseHealthChecker',
    'CacheHealthChecker'
]



from urllib.parse import urljoin

from django.urls import path

from health.api import resolve_health
from health.conf import configurations


def resolve_path(uri: str) -> str:
    """
    Resolves path by given uri.

    :param uri: uri
    :return: path with context path
    """
    return urljoin(configurations.context_path, uri)


urlpatterns = [
    path(resolve_path('health'), resolve_health)
]



import math
import typing
import importlib
import logging

from django.conf import ImproperlyConfigured

from health.consts import Status
from health.models import ComponentHealth
from health.checkers.base import HealthChecker


__all__ = ['checker', 'get_status']


log = logging.getLogger('health.service')


class _CheckerService:
    _checkers: typing.Dict[str, HealthChecker]

    def __init__(self):
        self._checkers = {}

    def register(self, checker_pkg: str) -> typing.NoReturn:
        """
        Register health checker instances.

        :param checker_pkg: checker_pkg
        """
        try:
            checker_module, checker_cls = checker_pkg.rsplit('.', 1)
            module = importlib.import_module(checker_module)
            cls = getattr(module, checker_cls)

            _checker = cls()
            if not isinstance(_checker, HealthChecker):
                raise ValueError(f"is not an instance of HealthChecker")
            _checker.setup()

            self._checkers[checker_pkg] = _checker

        except ModuleNotFoundError as ex:
            log.error(f'{checker_pkg} health checker: {ex}')
            raise ImproperlyConfigured('module not found')

        except AttributeError as ex:
            log.error(f'{checker_pkg} health checker: {ex}')
            raise ImproperlyConfigured('attribute error')

        except ValueError as ex:
            log.error(f'{checker_pkg} health checker: {ex}')
            raise ImproperlyConfigured('value error')

    def check(self) -> typing.List[ComponentHealth]:
        """
        Check all components

        :return: component health list
        """
        components = []
        for key, _checker in self._checkers.items():
            log.debug(f'Checking {key} health...')
            components += _checker.check()
        return components


def get_status(components: typing.List[ComponentHealth]) -> Status:
    """
    Calculates the over-all status by a given list of components health.

    :param components: list of health components
    :returns:
        Status.DEAD: if half or more than the half of the components have DEAD status
        Status.UNKNOWN: if half or more than the half of the components have UNKNOWN status
        Status.ALIVE: if more than the half of the components have ALIVE status
    """
    total = len(components)
    if total == 0:
        return Status.ALIVE

    components_alive = list(filter(lambda component: Status.ALIVE == component.status, components))
    components_dead = list(filter(lambda component: Status.DEAD == component.status, components))
    components_unknown = list(filter(lambda component: Status.UNKNOWN == component.status, components))

    half_total = math.floor(total/2.0)
    if len(components_dead) >= half_total:
        return Status.DEAD

    if len(components_unknown) >= half_total:
        return Status.UNKNOWN

    if len(components_alive) > half_total:
        return Status.ALIVE

    return Status.UNKNOWN


checker = _CheckerService()



import typing
import logging

from django.conf import settings
from django.db import connections
from django.db import Error as DBError

from health.checkers.base import HealthChecker
from health.consts import Status
from health.models import ComponentHealth
from health.models import DatabaseHealth


log = logging.getLogger('health.checkers.database')


class DatabaseHealthChecker(HealthChecker):
    """
    Database health checker.

    Executes only one query (SELECT 1) to check if the database is alive.
    """
    _databases: dict

    def setup(self) -> typing.NoReturn:
        self._databases = getattr(settings, 'DATABASES', {})

    def _get_uri(self, alias: str, vendor: str) -> str:
        db_conf = self._databases[alias]

        if vendor in ['sqlite', 'unknown']:
            return db_conf.get('NAME', '!unknown')

        host = db_conf.get('HOST', '!unknown')
        port = db_conf.get('PORT', '!unknown')
        uri = f"{vendor}://{host}:{port}"

        database = db_conf.get('NAME')
        if database:
            uri += f'/{database}'

        return uri

    def check(self) -> typing.List[ComponentHealth]:
        components = []
        for alias in self._databases:
            connection = connections[alias]

            health = DatabaseHealth()
            health.name = alias
            health.vendor = connection.vendor
            health.uri = self._get_uri(connection.vendor, alias)

            try:
                with connection.cursor() as cursor:
                    cursor.execute("SELECT 1")
                    one = cursor.fetchone()[0]
                    if one != 1:
                        raise AssertionError(f'Expected 1, got: {one}')

            except DBError as ex:
                log.error(f'DatabaseHealthChecker: {ex}')
                health.status = Status.DEAD

            except AssertionError as ex:
                log.warning(f'DatabaseHealthChecker: {ex}')
                health.status = Status.UNKNOWN

            finally:
                components.append(health)

        return components



from base import HealthChecker
from base import DisabledHealthChecker
from cache import CacheHealthChecker
from disk import DiskHealthChecker
from database import DatabaseHealthChecker

__all__ = [
    'HealthChecker', 'DisabledHealthChecker', 'DiskHealthChecker', 'DatabaseHealthChecker', 'CacheHealthChecker'
]



import typing

from health.models import ComponentHealth


class HealthChecker:
    def setup(self) -> typing.NoReturn:
        """
        Setup configurations before start.
        """
        pass

    def check(self) -> typing.List[ComponentHealth]:
        """
        Check component health.

        :return:
        """
        raise NotImplementedError()


class DisabledHealthChecker(HealthChecker):
    """
    Disabled health checker
    """
    def check(self) -> typing.List[ComponentHealth]:
        return []



import logging

from django.http.request import HttpRequest
from django.http.response import HttpResponse
from django.http.response import JsonResponse
from django.views.decorators import http

from health.service import checker
from health.service import get_status


__all__ = ['resolve_health']


log = logging.getLogger('health.api')


@http.require_GET
def resolve_health(request: HttpRequest) -> HttpResponse:
    """
    Endpoint to resolve health information.

    :param request: http request
    :returns: app health information and its components
    """
    log.info('Resolving health for application.')

    components = checker.check()

    return JsonResponse({
        "status": get_status(components),
        "components": [component.dict(exclude_unset=True) for component in components]
    }, status=200)



from django.apps import AppConfig


class HealthConfig(AppConfig):
    name = 'health'

    def ready(self):
        """
        Configure health checkers
        """
        from health.conf import configurations
        from health.service import checker

        if configurations.enabled:
            for checker_pkg in configurations.checkers:
                checker.register(checker_pkg)



import os
import shutil
import typing
import logging

from health.checkers.base import HealthChecker
from health.consts import Status
from health.models import ComponentHealth
from health.models import DiskHealth


log = logging.getLogger('health.checkers.disk')


class DiskHealthChecker(HealthChecker):
    _test_file_path: str
    _test_file_content: str

    def setup(self) -> typing.NoReturn:
        self._test_file_path = os.path.expanduser('~/disk_health_test.txt')
        self._test_file_content = 'HEALTHY'

    def check(self) -> typing.List[ComponentHealth]:
        health = DiskHealth()

        try:
            test = open(self._test_file_path, 'w')
            test.write(self._test_file_content)
            test.close()

            if not os.path.isfile(self._test_file_path):
                raise IOError(f'{self._test_file_path} is not a file')

            test = open(self._test_file_path, 'r')
            content = test.read()
            test.close()

            if self._test_file_content != content:
                raise AssertionError(f'Expected {self._test_file_content}, got: {content}')

            os.remove(self._test_file_path)
            if os.path.isfile(self._test_file_path):
                raise IOError(f'{self._test_file_path} was not removed')

            disk = shutil.disk_usage("/")

            health.total = disk.total
            health.free = disk.free
            health.used = disk.used

        except OSError as ex:
            log.error(f"DiskHealthChecker: {ex}")
            health.status = Status.DEAD

        except AssertionError as ex:
            log.warning(f"DiskHealthChecker: {ex}")
            health.status = Status.UNKNOWN

        finally:
            return [health]



from setuptools import setup


with open('README.md', 'r') as md:
    long_description = md.read()

setup(
  name='py-django-health',
  version='0.0.4',
  author='kencinas95',
  author_email='kencinas95@developers.com',
  description='Health checkers for Django apps',
  long_description=str(long_description),
  long_description_content_type='text/markdown',
  url='https://github.com/kencinas95/py-django-health',
  license='GPLv3',
  packages=['health', 'health.checkers'],
  keywords=['health', 'checkers', 'heartbeat', 'microservices'],
  install_requires=[
    'django',
    'pydantic'
  ],
  python_requires=">=3.6",
  classifiers=[
    'Development Status :: 3 - Alpha',
    'Intended Audience :: Developers',
    'Topic :: Software Development :: Build Tools',
    'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',
    'Programming Language :: Python :: 3',
    'Programming Language :: Python :: 3.4',
    'Programming Language :: Python :: 3.5',
    'Programming Language :: Python :: 3.6',
    'Programming Language :: Python :: 3.7',
    'Programming Language :: Python :: 3.8'
  ]
)



import typing
import logging

from django.conf import settings
from django.core.cache import caches
from django.core.cache import CacheKeyWarning

from health.checkers.base import HealthChecker
from health.consts import Status
from health.models import ComponentHealth
from health.models import CacheHealth


log = logging.getLogger('health.checkers.cache')


class CacheHealthChecker(HealthChecker):
    _caches: dict
    _test_key: str
    _test_content: str

    def setup(self) -> typing.NoReturn:
        self._caches = getattr(settings, 'CACHES', {})
        self._test_key = 'health.checker.cache.test.key'
        self._test_content = 'HEALTHY'

    def check(self) -> typing.List[ComponentHealth]:
        components = []
        for alias in getattr(settings, 'CACHES', {}):
            cache = caches[alias]

            health = CacheHealth()
            health.name = alias
            health.backend = self._caches[alias]['BACKEND']

            try:
                cache.set(self._test_key, self._test_content)

                content = cache.get(self._test_key)
                if content != self._test_content:
                    raise AssertionError(f'Expected {self._test_content}, got: {content}')

                cache.delete(self._test_key)
                if cache.get(self._test_key):
                    raise CacheKeyWarning(f'Key {self._test_key} was deleted but it is still there!')

            except AssertionError as ex:
                log.error(f'CacheHealthChecker: {ex}')
                health.status = Status.UNKNOWN

            except CacheKeyWarning as ex:
                log.warning(f'CacheHealthChecker: {ex}')
                health.status = Status.UNKNOWN

            except Exception as ex:
                log.error(f'CacheHealthChecker: {ex}')
                health.status = Status.DEAD

            finally:
                components.append(health)

        return components
